# stdlib
import os
import random

# third party
import ascii_magic

# relative
from .lib import asset_path


def motorcycle() -> None:
    print(
        """
                                             `
                                         `.+yys/.`
                                       ``/NMMMNNs`
                                    `./shNMMMMMMNs``    `..`
                                  `-smNMMNNMMMMMMN/.``......`
                                `.yNMMMMNmmmmNNMMm/.`....`
                              `:sdNMMMMMMNNNNddddds-`.`` `--. `
                           `.+dNNNNMMMMMMMMMNNNNmddohmh//hddy/.```..`
                          `-hNMMMMMMMMMMMMNNdmNNMNNdNNd:sdyoo+/++:..`
                        ../mMMMMMMMMMMMMMMNNmmmmNMNmNNmdmd/hNNNd+:`
                        `:mMNNMMMMMMMMMMMMNMNNmmmNNNNNdNNd/NMMMMm::
                       `:mMNNNMMMMMMMMMMMMMMMNNNNdNMNNmmNd:smMMmh//
                     ``/mMMMMMMMMMMMMMMMMMMMMMMNmdmNNMMNNNy/osoo/-`
                    `-sNMMMMMMMMMMMMMMMMMMMMMMMMNNmmMMMMNh-....`
                   `/dNMMMMMMMMMMMMMMMMMMMMMMMMMMMNNMMMNy.`
                ``.omNNMMMMMMMMMMMMNMMMMMMMNmmmmNNMMMMN+`
                `:hmNNMMMMMMMMMMMNo/ohNNNNho+os+-+hNys/`
                -mNNNNNNMMMMMMMMm+``-yNdd+/mMMMms.-:`
                .+dmNNNNMMMMMMNd:``:dNN+y`oMMMMMm-.`
                `+dmmmNNNmmmmy+.   `-+m/s/+MMMMm/--
               `+mmmhNy/-...```     ``-.-sosyys+/-`
            ``.smmmsoo``               .oh+-:/:.
          `.:odmdh/````             `.+d+``````
     ```/sydNdhy+.`              ``-sds.
    `:hdmhs::-````               `oNs.`
```.sdmh/``                    `-ym+`
 ``ssy+`                     `-yms.`
   ``                      `:hNy-``
   `                     `-yMN/```
                       `-yNhy-
                     `/yNd/`
                   `:dNMs``
                 `.+mNy/.`
              `.+hNMMs``
             `:dMMMMh.`"""  # noqa: W605
    )


def hold_on_tight() -> None:
    out = os.popen("stty size", "r").read().split()
    if len(out) == 2:
        rows, columns = out
    else:
        """not running in a proper command line (probably a unit test)"""
        return

    if int(columns) >= 91:

        print(
            """
 _   _       _     _                 _   _       _     _     _   _                       _
| | | |     | |   | |               | | (_)     | |   | |   | | | |                     | |
| |_| | ___ | | __| |   ___  _ __   | |_ _  __ _| |__ | |_  | |_| | __ _ _ __ _ __ _   _| |
|  _  |/ _ \| |/ _` |  / _ \| '_ \  | __| |/ _` | '_ \| __| |  _  |/ _` | '__| '__| | | | |
| | | | (_) | | (_| | | (_) | | | | | |_| | (_| | | | | |_  | | | | (_| | |  | |  | |_| |_|
\_| |_/\___/|_|\__,_|  \___/|_| |_|  \__|_|\__, |_| |_|\__| \_| |_/\__,_|_|  |_|   \__, (_)
                                            __/ |                                   __/ |
                                           |___/                                   |___/
            """  # noqa: W605
        )
    else:
        print(
            """
 _   _       _     _                 _   _                       _
| | | |     | |   | |               | | | |                     | |
| |_| | ___ | | __| |   ___  _ __   | |_| | __ _ _ __ _ __ _   _| |
|  _  |/ _ \| |/ _` |  / _ \| '_ \  |  _  |/ _` | '__| '__| | | | |
| | | | (_) | | (_| | | (_) | | | | | | | | (_| | |  | |  | |_| |_|
\_| |_/\___/|_|\__,_|  \___/|_| |_| \_| |_/\__,_|_|  |_|   \__, (_)
                                                            __/ |
                                                           |___/
        """  # noqa: W605
        )


def hagrid1() -> None:
    try:
        ascii_magic.to_terminal(
            ascii_magic.from_image_file(
                img_path=str(asset_path()) + "/img/hagrid.png", columns=83
            )
        )
    except Exception:
        pass


def hagrid2() -> None:
    try:
        ascii_magic.to_terminal(
            ascii_magic.from_image_file(
                img_path=str(asset_path()) + "/img/hagrid2.png", columns=83
            )
        )
    except Exception:
        pass


def hagrid() -> None:
    """Print a random hagrid image with the caption "hold on tight harry" """
    options = [motorcycle, hagrid1, hagrid2]
    i = random.randint(0, 2)
    options[i]()
    hold_on_tight()



"""The purpose of these functions is to check the local dependencies of the person running the CLI
tool and ensure that things are properly configured for the cli's full use (depending on the user's
operating system.) When dependencies are missing the CLI tool should offer helpful hints about what
course of action to take to install missing dependencies, even offering to run appropriate
installation commands where applicable."""

# stdlib
import platform
import shutil
import subprocess
from typing import Any
from typing import Dict
from typing import Optional


class MissingDependency(Exception):
    pass


def get_environment() -> Dict[str, Any]:
    return {
        "uname": platform.uname(),
        "platform": platform.system().lower(),
        "os_version": platform.release(),
        "python_version": platform.python_version(),
    }


ENVIRONMENT = get_environment()


def is_windows() -> bool:
    if "platform" in ENVIRONMENT and ENVIRONMENT["platform"] == "windows":
        return True
    return False


allowed_hosts = ["docker", "vm", "azure", "aws", "gcp"]
commands = ["docker", "git", "ansible-playbook"]

if is_windows():
    commands.append("wsl")


def check_deps() -> Dict[str, Optional[str]]:
    paths = {}
    for dep in commands:
        paths[dep] = shutil.which(dep)
    return paths


DEPENDENCIES = check_deps()


def docker_info() -> str:
    try:
        cmd = "docker info"
        output = subprocess.check_output(cmd, shell=True)
        return str(output.decode("utf-8"))
    except Exception as e:
        print("failed to get docker info", e)
        return str(e)


def wsl_info() -> str:
    try:
        cmd = "wsl --status"
        output = subprocess.check_output(cmd, shell=True)
        return str(output.decode("utf-8"))
    except Exception as e:
        print("failed to get wsl info", e)
        return str(e)


def wsl_linux_info() -> str:
    try:
        cmd = "wsl bash -c 'lsb_release -a'"
        output = subprocess.check_output(cmd, shell=True)
        return str(output.decode("utf-8"))
    except Exception as e:
        print("failed to get wsl linux info", e)
        return str(e)



# stdlib
import json
import os
from typing import Any

STABLE_BRANCH = "0.6.0"
DEFAULT_BRANCH = "dev"

arg_defaults = {
    "repo": "OpenMined/PySyft",
    "branch": STABLE_BRANCH,
    "username": "root",
    "auth_type": "key",
    "key_path": "~/.ssh/id_rsa",
    "azure_repo": "OpenMined/PySyft",
    "azure_branch": STABLE_BRANCH,
    "azure_username": "azureuser",
    "azure_key_path": "~/.ssh/id_rsa",
    "azure_resource_group": "openmined",
    "azure_location": "westus",
    "azure_size": "Standard_D4s_v3",
    "gcp_zone": "us-central1-c",
    "gcp_machine_type": "e2-standard-4",
    "gcp_project_id": "",
    "gcp_username": "",
    "gcp_key_path": "~/.ssh/google_compute_engine",
    "gcp_repo": "OpenMined/PySyft",
    "gcp_branch": STABLE_BRANCH,
}


class ArgCache:
    @staticmethod
    def cache_file_path() -> str:
        dir_path = os.path.expanduser("~/.hagrid")
        os.makedirs(dir_path, exist_ok=True)
        return f"{dir_path}/cache.json"

    def __init__(self) -> None:
        cache = {}
        try:
            with open(ArgCache.cache_file_path(), "r") as f:
                cache = json.loads(f.read())
        except Exception:
            pass
        self.__dict__ = cache

    def __setattr__(self, key: str, value: Any) -> None:
        super(ArgCache, self).__setattr__(key, value)
        if not key.startswith("__"):
            with open(ArgCache.cache_file_path(), "w") as f:
                f.write(json.dumps(self.__dict__))

    def __getattr__(self, key: str) -> Any:
        if key not in self.__dict__ and key in arg_defaults:
            return arg_defaults[key]
        else:
            print(f"Can't find key {key} in ArgCache")
            super().__getattr__(key)  # type: ignore


arg_cache = ArgCache()



# coding=utf-8
# stdlib
import subprocess
from typing import Callable
from typing import List

# one liner to use bootstrap script:
# CMD: curl https://raw.githubusercontent.com/OpenMined/PySyft/dev/packages/hagrid/hagrid/win_bootstrap.py > win_bootstrap.py && python win_bootstrap.py # noqa
# Powershell is complaining about a utf-8 issue we need to fix, could be related to a
# bug with long lines in utf-8
# PS: $r = Invoke-WebRequest "https://raw.githubusercontent.com/OpenMined/PySyft/dev/packages/hagrid/hagrid/win_bootstrap.py" -UseBasicParsing; echo $r.Content > win_bootstrap.py; python win_bootstrap.py # noqa


class Requirement:
    def __init__(
        self, full_name: str, choco_name: str, detect: Callable, extras: str = ""
    ) -> None:
        self.full_name = full_name
        self.choco_name = choco_name
        self.detect = detect
        self.extras = extras

    def __repr__(self) -> str:
        return self.full_name


install_choco_pwsh = """
[System.Net.ServicePointManager]::SecurityProtocol = [System.Net.ServicePointManager]::SecurityProtocol -bor 3072;
Invoke-Expression ((New-Object System.Net.WebClient).DownloadString('https://community.chocolatey.org/install.ps1'));
"""

install_wsl2_pwsh = """
wsl --update; wsl --shutdown; wsl --set-default-version 2; wsl --install -d Ubuntu; wsl --setdefault Ubuntu;
"""

# add this to block powershell from existing for debugging
# Read-Host -Prompt string


def make_admin_cmd(admin_cmd: str) -> str:
    return (
        f"Start-Process PowerShell -Wait -Verb RunAs -ArgumentList "
        '"'
        "Set-ExecutionPolicy Bypass -Scope Process -Force; "
        f"{admin_cmd}; "
        '"'
    )


def where_is(binary: str, req: Requirement) -> bool:
    print(f"{req.full_name} - {binary}", end="", flush=True)
    found = path_where_is(binary)
    if not found:
        found = full_where_is(binary)
    if found:
        print(" √")
    else:
        print(" ×")
    return found


def path_where_is(binary: str) -> bool:
    try:
        cmds = ["where.exe", binary]
        output = subprocess.run(cmds, capture_output=True, cwd="C:\\")
        out = str(output.stdout.decode("utf-8")).split("\r\n")
        if binary in out[0]:
            return True
    except Exception as e:
        print("error", e)
        pass
    return False


def full_where_is(binary: str) -> bool:
    try:
        powershell_cmd = f"where.exe /R C:\ *.exe | findstr \\{binary}$"  # noqa: W605
        cmds = ["powershell.exe", "-Command", powershell_cmd]
        output = subprocess.run(cmds, capture_output=True, cwd="C:\\")
        out = str(output.stdout.decode("utf-8")).split("\r\n")
        if binary in out[0]:
            return True
    except Exception as e:
        print("error", e)
        pass
    return False


def exe(binary: str) -> Callable:
    def call(req: Requirement) -> bool:
        return where_is(binary=binary, req=req)

    return call


def detect_wsl2(req: Requirement) -> bool:
    print(f"{req.full_name} - wsl.exe ", end="")
    try:
        powershell_cmd = "wsl.exe --status"
        cmds = ["powershell.exe", "-Command", powershell_cmd]
        output = subprocess.run(cmds, capture_output=True)
        out = output.stdout.decode("utf-16")
        if "Default Distribution: Ubuntu" in out:
            pass
        if "Default Version: 2" in out:
            print(" √")
            return True
    except Exception as e:
        print("error", e)
        pass
    print(" ×")
    return False


requirements = []
requirements.append(
    Requirement(
        full_name="Windows Subsystem for Linux 2",
        choco_name="wsl2",
        detect=detect_wsl2,
    )
)
requirements.append(
    Requirement(
        full_name="Chocolatey Package Manager",
        choco_name="choco",
        detect=exe("choco.exe"),
    )
)
requirements.append(
    Requirement(
        full_name="Anaconda Individual Edition",
        choco_name="anaconda3",
        detect=exe("conda.exe"),
    )
)
requirements.append(
    Requirement(
        full_name="Git Version Control",
        choco_name="git",
        detect=exe("git.exe"),
    )
)
requirements.append(
    Requirement(
        full_name="Docker Desktop",
        choco_name="docker-desktop",
        detect=exe("docker.exe"),
    )
)


def install_elevated_powershell(full_name: str, powershell_cmd: str) -> None:
    try:
        input(
            f"\nInstalling {full_name} requires Administrator.\n"
            "When the UAC dialogue appears click Yes on the left.\n\n"
            "Press enter to start..."
        )
        powershell_cmds = ["-command", powershell_cmd]
        output = subprocess.run(
            ["powershell.exe"] + powershell_cmds, capture_output=True
        )
        _ = output.stdout.decode("utf-8")
    except Exception as e:
        print("failed", e)


def install_choco() -> None:
    return install_elevated_powershell(
        full_name="Chocolatey", powershell_cmd=make_admin_cmd(install_choco_pwsh)
    )


def install_wsl2() -> None:
    return install_elevated_powershell(
        full_name="WSL2", powershell_cmd=make_admin_cmd(install_wsl2_pwsh)
    )


def install_deps(requirements: List[Requirement]) -> None:
    package_names = []
    for req in requirements:
        package_names.append(req.choco_name)

    try:
        input(
            "\nInstalling packages requires Administrator.\n"
            "When the UAC dialogue appears click Yes on the left.\n\n"
            "Press enter to start..."
        )
        choco_args = f"choco.exe install {' '.join(package_names)} -y"
        powershell_cmds = ["-command", make_admin_cmd(choco_args)]
        output = subprocess.run(
            ["powershell.exe"] + powershell_cmds, capture_output=True
        )
        _ = str(output.stdout.decode("utf-8"))
    except Exception as e:
        print("failed", e)


def ask_install(requirement: Requirement) -> bool:
    val = input(f"Do you want to install {requirement.full_name} (Y/n): ")
    if "y" in val.lower():
        return True
    return False


def check_all(requirements: List[Requirement]) -> List[Requirement]:
    missing = []
    for req in requirements:
        if not req.detect(req):
            missing.append(req)
    return missing


def main() -> None:
    print("\nHAGrid Windows Dependency Installer")
    print("===================================\n")
    print("Searching your computer for:")
    missing_deps = check_all(requirements=requirements)

    if len(missing_deps) > 0:
        print("\nWe were unable to find the following dependencies:")
        print("-----------------------------------")
        for dep in missing_deps:
            print(f"{dep.full_name}")

    print("")
    desired = []
    choco_required = False
    wsl2_required = False
    for dep in missing_deps:
        if ask_install(dep):
            if dep.choco_name == "choco":
                choco_required = True
            elif dep.choco_name == "wsl2":
                wsl2_required = True
            else:
                desired.append(dep)
        elif dep.choco_name == "choco":
            print("You must install Chocolatey to install other dependencies")
            return

    if wsl2_required:
        install_wsl2()

    if choco_required:
        install_choco()

    if len(desired) > 0:
        install_deps(desired)

    print("")
    still_missing = check_all(requirements=missing_deps)
    if len(still_missing) > 0:
        print("We were still unable to find the following dependencies:")
        print("-----------------------------------")
        for dep in still_missing:
            print(f"{dep.full_name}")
        print("Please try again.")
    else:

        print("\nCongratulations. All done.")
        print("===================================\n")
        print("Now you can run HAGrid on Windows!")


if __name__ == "__main__":
    main()
